import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * Representation of a MEAPsoft segment/feature file.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */

public class FeatFile extends MEAPFile
{
    String featureDescription = "";
    Vector chunks;  // Vector of FeatChunks
    
    // regular expressions for parsing FEATFiles
    private static final Pattern commentPattern = Pattern.compile("#\\.*");
    private static final Pattern linePattern = Pattern.compile("\\s*([^#\\s]+)\\s*");

    public FeatFile(String fn)
    {
        filename = fn;

        chunks = new Vector(100, 0);
    }


    /**
     * Get a matrix (2D double array) of all of the features in all of
     * the chunks contained in this file.  Each column corresponds to
     * a single feature vector.  Index it as you would in Matlab
     * ([row][column] = [featdim][chunk]).
     */
    public double[][] getFeatures()
    {
        return getFeatures(null);
    }

    /**
     * Get a matrix (2D double array) of all of the features in all of
     * the chunks contained in this file.  Each column corresponds to
     * a single feature vector.  Index it as you would in Matlab
     * ([row][column] = [featdim][chunk]).
     */
    public double[][] getFeatures(int[] featdim)
    {
        // how many feature dimensions are we using?
        int maxdim = 0;
        if(featdim != null)
            maxdim = featdim.length;
        else
            maxdim = ((FeatChunk)chunks.get(0)).numFeatures();

        double[][] mat = new double[chunks.size()][maxdim];

        for(int x = 0; x < chunks.size(); x++)
        {
            FeatChunk c = (FeatChunk)chunks.get(x);

            double[] currFeat = c.getFeatures(featdim);
            for(int y = 0; y < currFeat.length; y++)
                mat[x][y] = currFeat[y];
        }
        
        return mat;
    }

    /**
     * Read in an feature file
     */
    public void readFile() throws IOException, ParserException
    {
        BufferedReader in  = new BufferedReader(new FileReader(filename));

        String audioFile;
        double chunkStartTime;
        double chunkLength;

        // each line (excluding comments) should look like: 
        // audioFile chunkStartTime chunkLength feature1 feature2 ...

        // Parse each line of the input file            
        String header = "# Features: ";
        boolean haveWrittenHeader = false;
        long lineno = 0;
        String line;
        while((line = in.readLine()) != null) 
        { 
            lineno++;
            
            // extract any comments from the current line
            String comment = "";
            Matcher c = commentPattern.matcher(line+"\n");
            if(c.find())
            {
                // comments go all the way to the end of the line
                comment = c.group() + line.substring(c.end()) + "\n";
                line = line.substring(0, c.start());
            }

            Matcher p = linePattern.matcher(line);
            // is there anything else?
            if(!p.find())
                continue;
            audioFile = p.group(1);
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk start time.");
            try { chunkStartTime = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk start time \"" 
                                          + p.group(1)  + "\".");  }
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk length.");
            try { chunkLength = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk length \"" 
                                          + p.group(1)  + "\".");  }
            
            FeatChunk ch = new FeatChunk(audioFile, chunkStartTime, chunkLength);
            ch.comment = comment;

            // save the remaining features on the line
            while(p.find())
            {
                // what kind of feature is this?  If its not a double then its a string;
                try 
                {
                    ch.addFeature(Double.parseDouble(p.group(1)));
                }
                catch (NumberFormatException e)
                {
                    ch.addFeature(p.group(1));
                }
            }

            chunks.add(ch);

            if(comment.startsWith("# Features: "))
            {
                featureDescription = comment.substring(12);
            }
        }

        in.close();
    }
    
    /**
     * Write the contents of this FeatFile
     */
    protected void write(Writer w) throws IOException
    {
        // write the header
        w.write("# filename onset_time chunk_length [features]\n# Features: ");
        w.write(featureDescription + "\n");

        Iterator i = chunks.iterator();
        while(i.hasNext())
            w.write(i.next().toString());
    }
}
