import java.io.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import java.text.*;

/**
 * A panel containing a graphical visualization of a matrix of data.
 * Like Matlab's imagesc() command
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class ImagePanel extends JPanel implements ActionListener, MouseMotionListener, MouseListener
{
    // constants
    private static final float VZOOM_INCR = 2;
    private static final float HZOOM_INCR = 2;
    // Actions performed by various buttons
    private static final String ZOOM_RESET = "zoom_reset";
    private static final String ZOOM_IN = "zoom_in";
    private static final String ZOOM_OUT = "zoom_out";
    private static final String ZOOM_V = "zoom_v";
    private static final String ZOOM_H = "zoom_h";

    FeatFile featFile;
    int[] featdim;

    SpectrogramPanel dataPanel;
    JScrollPane scroller;
    JLabel status;
    JToolBar toolbar;
    JToggleButton button_zoom_in;
    JToggleButton button_zoom_out;

    // states to keep track of
    float currHZoom = 1.0f;
    float currVZoom = 1.0f;
    // are we in the zoomIn state or zoomOut? flag controlled by zoom
    // buttons on the menu bar
    boolean zoomIn = true;
    boolean hZoom = true;
    boolean vZoom = false;
    // keep track of where zoom started when dragging a zoombox
    Point zoomStart = null;
    
    public ImagePanel(String filename)
    {
        this(filename, null);
    }

    public ImagePanel(FeatFile file)
    {
        this(file, null);
    }

    public ImagePanel(String filename, int[] fd)
    {
        this(new FeatFile(filename), fd);
    }

    public ImagePanel(FeatFile file, int[] fd )
    {
        super();
        featFile = file;
        featdim = fd;
    }

    public void setup() throws IOException, ParserException
    {
        featFile.readFile();

        dataPanel = new SpectrogramPanel(featFile.getFeatures(featdim));

        setLayout(new BorderLayout());

        //TODO:
        // _ still need to set up axes.  and colorbar zooming
        // X want status bar on the bottom that lists value in data matrix
        //   at point indicated by mouse location 
        // X want to be able to do proper vertical scaling (automatically
        //   fill the window unless vertical scroll is set manually (then
        //   let scrollbar take over)
        // X zooming and scrolling don't play nice together yet -  Fixed in SpectrogramPanel.zoom 
        // C need to find out how to get the current visible size of the
        //   window and set up actionlisteners/whatever to update things

	    scroller = new JScrollPane(dataPanel);
        add(scroller, BorderLayout.CENTER);
        //add(hAxis, BorderLayout.WEST);
        //add(dataPanel.getColorBar(), BorderLayout.EAST);


        // tool bar
        toolbar = new JToolBar(JToolBar.HORIZONTAL);
        add(toolbar, BorderLayout.PAGE_START);
        // keep it anchored at the top
        toolbar.setFloatable(false);

        // toolbar buttons
        toolbar.add(new JLabel("Zoom:"));
        toolbar.addSeparator();

        JToggleButton tbutton = new JToggleButton("h", hZoom);
        tbutton.setActionCommand(ZOOM_H);
        tbutton.addActionListener(this);
        toolbar.add(tbutton);

        tbutton = new JToggleButton("v", vZoom);
        tbutton.setActionCommand(ZOOM_V);
        tbutton.addActionListener(this);
        toolbar.add(tbutton);
        toolbar.addSeparator();

        tbutton = new JToggleButton("in", zoomIn);
        tbutton.setActionCommand(ZOOM_IN);
        tbutton.addActionListener(this);
        toolbar.add(tbutton);
        button_zoom_in = tbutton;

        tbutton = new JToggleButton("out", !zoomIn);
        tbutton.setActionCommand(ZOOM_OUT);
        tbutton.addActionListener(this);
        toolbar.add(tbutton);
        button_zoom_out = tbutton;
        toolbar.addSeparator();

        JButton button = new JButton("reset");
        button.setActionCommand(ZOOM_RESET);
        button.addActionListener(this);
        toolbar.add(button);


        // status bar
        status = new JLabel("");
        add(status, BorderLayout.SOUTH);
        dataPanel.addMouseMotionListener(this);

        // Mouse listener to zoom in/out on mouse clicks
        dataPanel.addMouseListener(this);

        // set the default size and reset the display
        this.setPreferredSize(new Dimension(500, 300));
        actionPerformed(new ActionEvent(button, 0, ZOOM_RESET));
    }


    public void actionPerformed(ActionEvent e) 
    {
        //String cmd = ((AbstractButton)e.getSource()).getActionCommand();
        String cmd = e.getActionCommand();
            
        if(cmd.equals(ZOOM_IN))
        {
            zoomIn = true;
            button_zoom_in.setSelected(true); 
            button_zoom_out.setSelected(false); 
        }
        else if(cmd.equals(ZOOM_OUT))
        {
            zoomIn = false;
            button_zoom_in.setSelected(false);
            button_zoom_out.setSelected(true);
        }
        else if(cmd.equals(ZOOM_RESET))
        {
            float h = (float)(this.getPreferredSize().getHeight() 
                              - scroller.getHorizontalScrollBar()
                                        .getPreferredSize().getHeight());
            // no, this should not be hardcoded (the 3 especially), but I
            // don't know how else to properly set the initial zoom so
            // that no vertial scrollbar is required
            boolean vz = vZoom, hz = hZoom;
            vZoom = true;  hZoom = true;
            zoomDataPanel(1, (h-3)/(float)dataPanel.getDataHeight());
            vZoom = vz;  hZoom = hz;
        }
        else if(cmd.equals(ZOOM_V))
            vZoom = !vZoom;
        else if(cmd.equals(ZOOM_H))
            hZoom = !hZoom;
    }

    public void mouseMoved(MouseEvent e)
    { 
        int x = (int)(e.getX()/currHZoom);
        // y axis is flipped (axis xy) 
        int y = (int)dataPanel.getDataHeight()-(int)(e.getY()/currVZoom);

        if(x < dataPanel.getDataWidth() && x >= 0 
           && y < dataPanel.getDataHeight() && y >= 0)   
            status.setText(featFile.filename + ": features(" + x + "," + y 
                           + ") = " + dataPanel.getData(x,y));
    }

    public void mouseDragged(MouseEvent e) { }

    public void mouseClicked(MouseEvent e) 
    { 
        boolean zoomin = zoomIn && e.getButton() == MouseEvent.BUTTON1 
            || !zoomIn && e.getButton() == MouseEvent.BUTTON3;
        if(zoomin)
            zoomDataPanel(currHZoom*HZOOM_INCR, currVZoom*VZOOM_INCR);
        else if(!zoomin)
            zoomDataPanel(currHZoom/HZOOM_INCR, currVZoom/VZOOM_INCR);
    }

    // take care of dragging a zoom rectangle and whatnot here
    public void mousePressed(MouseEvent e) { }
    public void mouseReleased(MouseEvent e) { }

    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }


    private void zoomDataPanel(float h, float v)
    {
        if(hZoom)
        {
            currHZoom = h;
            dataPanel.hzoomSet(currHZoom);
        }
        if(vZoom)
        {
            currVZoom = v;
            dataPanel.vzoomSet(currVZoom);
        }
    }

    public static void main(String[] args)
    {
        // parse arguments
        int[] featdim = MEAPUtil.parseFeatDim(args,"i:"); 
        String filename = args[args.length-1];
        
        ImagePanel p = new ImagePanel(filename, featdim);

        try
        { 
            p.setup();
        }
        catch(Exception e)
        {
            e.printStackTrace();
            System.exit(1);
        }

        JFrame jframe = new JFrame(p.getClass().getName() + ": " + filename);
        jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        jframe.setContentPane(p);
        jframe.pack();
        jframe.setVisible(true);
    }
}
