import javax.sound.sampled.*;
import java.io.*;
import gnu.getopt.*;

import featextractors.*;

/**
 * Program to segment input files and output text feature files
 * compatible with other MEAPsoft tools.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */

public class Segmenter extends MEAPUtil
{
    // Files to use
    String[] audioFiles;
    String outFileName="out.segments";
    FeatFile outFile; 
    
    // Extractor parts
    AudioInputStream ais;
    STFT stft;
    boolean useBeatOD = true;
    OnsetDetector detector;
    SegmentExtractor extractor;
    
    boolean onsetAtFirstFrame = false;
    
    public void printUsageAndExit() 
    {
        System.out.println("Usage: Segmenter [-options] source1.wav source2.mp3 ... \n\n" + 
               "  where options include:\n" + 
		       "    -o output_file  the file to write the output to (defaults to ./out.segments)\n" +
		       "    -m N      [0]   use mixer N for input\n" +
		       "    -b D.D  [2.0]   fraction of the band mean needed to trigger a band onset\n" +
		       "    -f D.D   [.3]   fraction of bands that need to onset for a global onset\n" +
               "    -d              use the old style onset detector (defaults to BeatOnsetDetector)\n" +
               "    -0              add an onset at the very beginning of the file\n" +
		       "");
        System.exit(0);
    }

    public Segmenter(String[] inFiles, String outputFile)
    {
        audioFiles = inFiles;

        outFileName = outputFile;
        outFile = new FeatFile(outFileName);
    }

    public Segmenter(String inFile, String outputFile)
    {
        audioFiles = new String[1];
        audioFiles[0] = inFile;

        outFileName = outputFile;
        outFile = new FeatFile(outFileName);
    }

    public Segmenter(String[] inFiles, String outputFile, int mixer, double bandThr, double bandFr, boolean beatOD)
    {
        this(inFiles, outputFile);

        mixerToUse = mixer;
        bandThresh = bandThr;
        bandFrac = bandFr;
        useBeatOD = beatOD;
    }

    public Segmenter(String[] args) 
    {
        if(args.length == 0)
            printUsageAndExit();
        
        // Parse arguments
        Getopt opt = new Getopt("Segmenter", args, "m:b:f:do:0");
        opt.setOpterr(false);
        
        int c = -1;
        while ((c = opt.getopt()) != -1) 
        {
            switch(c) 
            {
            case 'm':
                mixerToUse = Integer.parseInt(opt.getOptarg());
                System.out.println("Using mixer " + mixerToUse);
                break;
            case 'b':
                bandThresh = Double.parseDouble(opt.getOptarg());
                System.out.println("bandThresh = " + bandThresh);
                break;
            case 'f':
                bandFrac = Double.parseDouble(opt.getOptarg());
                System.out.println("bandFrac = " + bandFrac);
                break;
            case 'd':
                useBeatOD = false;
                break;
            case 'o':
                outFileName = opt.getOptarg();
                break;
            case '0':
                onsetAtFirstFrame = true;
                break;
            case '?':
                printUsageAndExit();
                break;
            default:
                System.out.print("getopt() returned " + c + "\n");
            }
        }
        
        // parse arguments
        int ind = opt.getOptind();
        if(ind > args.length)
            printUsageAndExit();
        
        audioFiles = new String[args.length - ind];
        for(int i=ind; i<args.length; i++)
            audioFiles[i-ind] = args[i];

        outFile = new FeatFile(outFileName);
    }

    public FeatFile processAudioFile(String fn)
    {
        // Create Extractor
        ais = openInputStream(fn);
        
        stft = new STFT(ais, frameSize, frameLatency);
        if(useBeatOD)
            detector = new BeatOnsetDetector(stft, bandThresh, bandFrac);
        else
            detector = new OnsetDetector(stft, bandThresh, bandFrac);
        extractor = new SegmentExtractor(stft, detector, fn, outFile);
        detector.addOnsetListener(extractor);

        // Hook things together
        stft.addFrameListener(extractor);
        
        // insert an onset at frame 0?
        if(onsetAtFirstFrame) 
            extractor.newOnset(0, 0);

        // Extract the entire source file
        extractor.run();

        return outFile;
    }
   
    public void go() 
    {
        // process supplied files
        for(int i=0; i<audioFiles.length; i++) 
        {
            System.out.println("Writing segments from " + audioFiles[i] + " to " 
                               + outFileName + ".");
            
            long startTime = System.currentTimeMillis();
            processAudioFile(audioFiles[i]);

            try 
            { 
                outFile.writeFile();
            }
            catch(IOException e)
            { 
                e.printStackTrace();
                System.exit(1);
            }
            
            System.out.println("Done.  Took " + 
                               ((System.currentTimeMillis() - startTime)/1000.0) 
                               + "s");
        }
    }

    public static void main(String[] args) 
    {
        Segmenter o2or = new Segmenter(args);
        o2or.go();
        System.exit(0);
    }
}
