import javax.sound.sampled.*;
import java.io.*;
import java.util.*;
import gnu.getopt.*;
import java.text.NumberFormat;

import featextractors.*;

/**
 * Program that composes an EDL by sorting chunks in featFile based on
 * some subset of their features using some distance metric.  Sorting
 * is done based on distance from the origin in feature space.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class SortComposer extends Composer
{
    String outFileName = "sorted.edl";

    FeatFile featFile;
    int[] featdim = null;
    ChunkDist dist;
    boolean reverseSort = false;
    boolean debug = false;

    public void printUsageAndExit() 
    {
        System.out.println("Usage: SortComposer [-options] features.feat \n\n" + 
               "  where options include:\n" + 
               "    -o output_file  the file to write the output to (defaults to sorted.edl)\n" +
               "    -g              debug mode (prints out chunk features on each line of output file)\n" +
               "    -r              sort in reverse order");
        printCommandLineOptions('i');
        printCommandLineOptions('d');
        System.out.println();
        System.exit(0);
    }


    public SortComposer(String featFN, String outFN, ChunkDist cd)
    {
        this(new FeatFile(featFN), new EDLFile(outFN), cd);
    }

    public SortComposer(FeatFile featFN, EDLFile outFN, ChunkDist cd)
    {
        featFile = featFN;
        outFile = outFN;
        dist = cd;
        featdim = cd.featdim;

        if(outFile == null)
            outFile = new EDLFile("");
    }

    public SortComposer(String[] args) 
    {
        if(args.length == 0)
            printUsageAndExit();

        Vector features = new Vector();

        // Parse arguments
        String argString = "d:i:o:rg";
        featdim = parseFeatDim(args, argString);
        dist = parseChunkDist(args, argString, featdim);

        Getopt opt = new Getopt("SortComposer", args, argString);
        opt.setOpterr(false);
        
        int c = -1;
        while ((c =opt.getopt()) != -1) 
        {
            switch(c) 
            {
            case 'o':
                outFileName = opt.getOptarg();
                break;
            case 'r':
                reverseSort = true;
                break;
            case 'g':
                debug = true;
                break;
            case 'd':  // already handled above
                break;
            case 'i':  // already handled above
                break;
            case '?':
                printUsageAndExit();
                break;
            default:
                System.out.print("getopt() returned " + c + "\n");
            }
        }
        
        // parse arguments
        int ind = opt.getOptind();
        if(ind > args.length)
            printUsageAndExit();
        
        featFile = new FeatFile(args[args.length-1]);
        outFile = new EDLFile(outFileName);

        System.out.println("Composing " + outFileName + 
                           " from " +  args[args.length-1] + ".");
    }

    public void setup()
    {
        try
        {
            featFile.readFile();
        }
        catch(Exception e)      
        {
            e.printStackTrace();
            System.exit(1);
        } 

        if(featFile.chunks.size() == 0)
        {
            System.out.println("Error: Empty feature file");
            System.exit(1);
        }
    }
    
    public EDLFile compose()
    {
        // how many feature dimensions are we using?
        int maxdim = 0;
        if(featdim != null)
        {
            for(int x = 0; x < featdim.length; x++)
            {
                if(featdim[x] > maxdim)
                    maxdim = featdim[x]; 
            }
        }
        else
            maxdim = ((FeatChunk)featFile.chunks.get(0)).numFeatures();
        
        FeatChunk targetChunk = new FeatChunk("",0,0);
        for(int i = 0; i <= maxdim; i++)
            targetChunk.addFeature(0);
        
        dist.setTarget(targetChunk);
        
        // maintain a set of chunks sorted using dist from targetChunk
        Heap chunks = null;
        if(reverseSort)
            chunks = new MaxHeap(dist);
        else
            chunks = new MinHeap(dist);
        chunks.addAll(featFile.chunks);
        
        NumberFormat fmt = NumberFormat.getInstance();
        fmt.setMaximumFractionDigits(3);
        
        double currTime = 0;
        while(chunks.size() > 0)
        {
            FeatChunk match = (FeatChunk)chunks.remove(0);
            
            // turn match chunk into an EDL chunk 
            EDLChunk nc = new EDLChunk(match, currTime);

            if(debug)
            {
                nc.comment = "    # feats = ";
                double[] feat = match.getFeatures(featdim);
                for(int x = 0; x < feat.length-1; x++)
                    nc.comment += fmt.format(feat[x]) + ", ";
                nc.comment += fmt.format(feat[feat.length-1]);
            }

            outFile.chunks.add(nc);
            
            currTime += match.length;
        }

        return outFile;
    } 

    public static void main(String[] args) 
    {
        SortComposer m = new SortComposer(args);
        long startTime = System.currentTimeMillis();
        m.go();
        System.out.println("Done. Took " +
                           ((System.currentTimeMillis() - startTime)/1000.0)
                           + "s");
        System.exit(0);
    }
}
