package featextractors;

import java.util.*;

/**
 * Averages all spectral frames together into a single feature vector
 * and then converts the vector to the mel frequency scale.
 *
 * @author Mike Mandel (mim@ee.columbia.edu)
 */

public class AvgMelSpec implements FeatureExtractor {

  double[] centersMel;
  int[] centersInd;
  double[] freq2mel;
  double[] linSpec;
  int N, outDim;

  public double lin2mel(double fq) {
    return 1127.0 * Math.log(1.0 + fq / 700.0);
  }

  public double mel2lin(double mel) {
    return 700.0 * (Math.exp(mel / 1127.0) - 1.0);
  }

  // Default constructor - default values are hardcoded... This is ugly.
  public AvgMelSpec() {
      this(129, 40, 44100.0);
  }

  public AvgMelSpec(int N, int outDim, double sampleRate) {
    this.N = N;
    this.outDim = outDim;
    linSpec = new double[N];

    // Calculate the locations of the bin centers on the mel scale and
    // as indices into the input vector
    centersMel = new double[outDim+2];
    centersInd = new int[outDim+2];
    double melMin = lin2mel(0);
    double melMax = lin2mel(1 / (2*sampleRate));
    for(int i=0; i<outDim+2; i++) {
      centersMel[i] = melMin + i * (melMax - melMin) / (outDim + 1);
      
      // Indices in inBuf at the smallest integer greater than the mel
      // triangle peaks
      centersInd[i] = (int) (mel2lin(centersMel[i])*N*sampleRate*2 + 1);
    }

    // precalculate mel translations of fft bin frequencies
    freq2mel = new double[N];
    for(int i=0; i<N; i++)
      freq2mel[i] = lin2mel(i / (2*N * sampleRate));
  }


  public double[] features(STFT stft, long startFrame, int length) {
    double[] melSpec = new double[outDim];
    double[] curFrame;
    double sum = 0;

    // intialize average to 0
    Arrays.fill(linSpec, 0);

    // collect average linear spectrum
    for(int frame=0; frame<length; frame++) {
      curFrame = stft.getFrame(startFrame+frame);
      for(int band=0; band<linSpec.length; band++)
        linSpec[band] += curFrame[band] / length;
    }

    // convert log magnitude to linear magnitude for binning
    for(int band=0; band<linSpec.length; band++)
      //linSpec[band] = Math.exp(linSpec[band]);
      linSpec[band] = Math.pow(10,linSpec[band]/10);

    // convert to mel scale
    for(int bin=0; bin<outDim; bin++) {
      // initialize
      melSpec[bin] = 0;
      
      // ramp up
      for(int i=centersInd[bin]; i<centersInd[bin+1]; i++) 
	melSpec[bin] += linSpec[i] * (freq2mel[i] - centersMel[bin]) 
	  / (centersMel[bin+1] - centersMel[bin]);
      
      // ramp down
      for(int i = centersInd[bin+1]; i < centersInd[bin+2]; i++)
	melSpec[bin] += linSpec[i] * (centersMel[bin+2] - freq2mel[i]) 
	  / (centersMel[bin+2] - centersMel[bin+1]);

      // Take log
      melSpec[bin] = 10*Math.log(melSpec[bin]) / Math.log(10);

      sum += melSpec[bin];
    }

    // Audio scrubber takes care of normalization, level is a good cue
//     for(int bin=0; bin<outDim; bin++)
//       melSpec[bin] = melSpec[bin] / sum;
    
    return melSpec;
  }
}
